const sliceType = 1;

const rootType = 2;
const selectorType = 3;
const selectorAllType = 4;
const parentType = 5;
const firstChildType = 6;
const lastChildType = 7;
const prevSiblingType = 8;
const nextSiblingType = 9;

const removeType = 10;
const clearType = 11;

const htmlType = 12;
const replaceType = 13;
const appendType = 14;
const prependType = 15;
const insertAfterType = 16;
const insertBeforeType = 17;

const setAttrType = 18;
const replaceAttrType = 19;
const rmAttrType = 20;
const setStylesType = 21;
const rmStylesType = 22;
const addClassesType = 23;
const rmClassesType = 24;

function factory(type) {
  return function() {
    var result = [type];

    for (let i = 0;i < arguments.length;i++) {
      result.push(arguments[i]);
    }

    return result
  }
}

const nil = [0];
const remove = [removeType];
const clear = [clearType];

const list =           factory(sliceType);
const root =           factory(rootType);
const one =            factory(selectorType);
const all =            factory(selectorAllType);
const parent =         factory(parentType);
const firstChild =     factory(firstChildType);
const lastChild =      factory(lastChildType);
const prevSibling =    factory(prevSiblingType);
const nextSibling =    factory(nextSiblingType);
const html =           factory(htmlType);
const replace =        factory(replaceType);
const append =         factory(appendType);
const prepend =        factory(prependType);
const insertAfter =    factory(insertAfterType);
const insertBefore =   factory(insertBeforeType);
const setAttr =        factory(setAttrType);
const replaceAttr =        factory(replaceAttrType);
const rmAttr =         factory(rmAttrType);
const setStyles =      factory(setStylesType);
const rmStyles =       factory(rmStylesType);
const addClasses =       factory(addClassesType);
const rmClasses =        factory(rmClassesType);

function mapNodes(nodes, mapFn) {
  var result = [];
  var children, i, j, c;

  for (i = 0;i < nodes.length;i++) {
    children = mapFn(nodes[i]);
    for (j = 0;j < children.length;j++) {
      c = children[j];
      if (c && c.nodeType == 1) {
        result.push(c);
      }
    }
  }

  return result
}

function mapQuerySelector(selector) {
  return node => [node.querySelector(selector)]
}

function mapQuerySelectorAll(selector) {
  return node => node.querySelectorAll(selector)
}

function mapParent(node) {
  return [node.parentNode]
}

function mapFirstChild(node) {
  var i, n, cn = node.childNodes;

  for (i = 0;i < cn.length;i++) {
    n = cn[i];
    if (n.nodeType==1) {
      return [n]
    }
  }

  return []
}

function mapLastChild(node) {
  var i, n, cn = node.childNodes;

  for (i = cn.length - 1;i >= 0;i--) {
    n = cn[i];
    if (n.nodeType==1) {
      return [n]
    }
  }

  return []
}

function mapPrevSibling(node) {
  var ps = node.previousSibling;
  while (ps && ps.nodeType !== 1) ps = ps.previousSibling;
  return ps ? [ps] : []
}

function mapNextSibling(node) {
  var ns = node.nextSibling;
  while (ns && ns.nodeType !== 1) ns = ns.nextSibling;
  return ns ? [ns] : []
}

function fromClass(className) {
  var classMap = {};
  var classes = className.split(/[\s\uFEFF\xA0]+/);
  var i, c;

  for (i = 0;i < classes.length;i++) {
    c = classes[i];
    if (c) {
      classMap[c] = true;
    }
  }

  return classMap
}

function toClass(classMap) {
  var classNames = [];
  var i;

  for (i in classMap) {
    if (classMap.hasOwnProperty(i) && classMap[i]) {
      classNames.push(i);
    }
  }

  return classNames.join(' ')
}

function setAttribute(node, key, value) {
  node.setAttribute(key, value);
  if (key == 'value' && node.tagName.toLowerCase() == 'input') {
    node.value = value;
  }
}

function forEach(elems, operation) {
  return new Promise(resolve => {
    let i = -1;

    function step() {
      i++;
      if (i < elems.length) {
        operation(elems[i]).then(step, step);
      } else {
        resolve();
      }
    }

    step();
  })
}

function clone(elems) {
  const result = [];
  for (let i = 0;i < elems.length;i++) {
    result.push(elems[i]);
  }

  return result
}

function getFragment(node, html) {
  const fragment = node.cloneNode();
  fragment.innerHTML = html;
  markSingularNodes(fragment);
  return fragment
}

function mapScript(node) {
  const isAttr = node.getAttribute('is');
  const document = node.ownerDocument;

  let script;
  if (isAttr) {
    script = document.createElement('script', { is: isAttr });
  } else {
    script = document.createElement('script');
  }

  script.text = node.text;
  for (let i = 0; i < node.attributes.length; i++) {
    const attribute = node.attributes[i];
    script.setAttribute(attribute.name, attribute.value);
  }

  return script
}

function onLoad(node) {
  return new Promise(resolve => {
    if (!('onload' in node)) {
      resolve();
      return
    }

    if (node.addEventListener) {
      node.addEventListener('load', resolve, false);
      node.addEventListener('error', resolve, false);
      return
    }

    if (node.attachEvent) {
      node.attachEvent('onload', resolve);
      node.attachEvent('onerror', resolve);
      node.attachEvent('onreadystatechange', () => {
        if (node.readyState == 'loaded' || node.readyState == 'complete') {
          resolve();
        }
      });

      return
    }

    resolve();
  })
}

function markSingularNodes(fragment) {
  const singularNodes = fragment.querySelectorAll('script, link[rel=stylesheet]');

  for (let i = 0;i < singularNodes.length;i++) {
    let node = singularNodes[i];

    while (node) {
      node.__witSingularNode = true;
      node = node.parentNode;
    }
  }
}

function waitUntilScriptLoaded(node) {
  if (node.src) {
    return onLoad(node)
  }

  return Promise.resolve()
}

function waitUntilLinkLoaded(node) {
  if (node.href && node.rel == 'stylesheet') {
    return onLoad(node)
  }

  return Promise.resolve()
}

function transfer(trf, parent) {
  return forEach(clone(parent.childNodes), (node) => {
    if (!node.__witSingularNode) {
      trf(node);
      return Promise.resolve()
    }

    switch (node.tagName.toLowerCase()) {
      case 'script':
        const script = mapScript(node);
        trf(script);
        return waitUntilScriptLoaded(script)
      case 'link':
        trf(node);
        return waitUntilLinkLoaded(node)
      default:
        const clone = node.cloneNode();
        trf(clone);
        return transfer(child => clone.appendChild(child), node)
    }
  })
}

function apply(delta, rootNode, nodes) {

  switch (delta[0]) {

    case sliceType:
      return forEach(delta.slice(1), d => apply(d, rootNode, nodes))

    case rootType:
      return forEach(delta.slice(1), d => apply(d, rootNode, [rootNode]))

    case selectorType:
      return forEach(delta.slice(2), d => apply(d, rootNode, mapNodes(nodes, mapQuerySelector(delta[1]))))

    case selectorAllType:
      return forEach(delta.slice(2), d => apply(d, rootNode, mapNodes(nodes, mapQuerySelectorAll(delta[1]))))

    case parentType:
      return forEach(delta.slice(1), d => apply(d, rootNode, mapNodes(nodes, mapParent)))

    case firstChildType:
      return forEach(delta.slice(1), d => apply(d, rootNode, mapNodes(nodes, mapFirstChild)))

    case lastChildType:
      return forEach(delta.slice(1), d => apply(d, rootNode, mapNodes(nodes, mapLastChild)))

    case prevSiblingType:
      return forEach(delta.slice(1), d => apply(d, rootNode, mapNodes(nodes, mapPrevSibling)))

    case nextSiblingType:
      return forEach(delta.slice(1), d => apply(d, rootNode, mapNodes(nodes, mapNextSibling)))

    case removeType: {
      for (let i = 0;i < nodes.length;i++) {
        const node = nodes[i];
        if (node.parentNode) {
          node.parentNode.removeChild(node);
        }
      }

      break
    }

    case clearType: {
      for (let i = 0;i < nodes.length;i++) {
        nodes[i].innerHTML = '';
      }

      break
    }

    case setAttrType: {
      const attrMap = delta[1];
      for (let i = 0;i < nodes.length;i++) {
        const node = nodes[i];
        for (let j in attrMap) if (attrMap.hasOwnProperty(j)) {
          setAttribute(node, j, attrMap[j]);
        }
      }

      break
    }

    case replaceAttrType: {
      const attrMap = delta[1];
      for (let i = 0;i < nodes.length;i++) {
        const node = nodes[i];

        for (let j = node.attributes.length-1;j >= 0;j--) {
          const name = node.attributes[j].name;
          if (!attrMap.hasOwnProperty(name)) {
            node.removeAttribute(name);
          }
        }

        for (let j in attrMap) if (attrMap.hasOwnProperty(j)) {
          setAttribute(node, j, attrMap[j]);
        }
      }

      break
    }

    case rmAttrType: {
      for (let i = 0;i < nodes.length;i++) {
        const node = nodes[i];
        for (let j = 1;j < delta.length;j++) {
          node.removeAttribute(delta[j]);
        }
      }

      break
    }

    case setStylesType: {
      const stylesMap = delta[1];
      for (let i = 0;i < nodes.length;i++) {
        const node = nodes[i];

        for (let j in stylesMap) if (stylesMap.hasOwnProperty(j)) {
          const m = stylesMap[j].match(/^(.*?)(\s*)!important$/);
          if (m) {
            node.style.setProperty(j, m[1], 'important');
          } else {
            node.style.setProperty(j, stylesMap[j]);
          }
        }
      }

      break
    }

    case rmStylesType: {
      for (let i = 0;i < nodes.length;i++) {
        const node = nodes[i];
        for (let j = 1;j < delta.length;j++) {
          node.style.removeProperty(delta[j]);
        }
      }

      break
    }

    case addClassesType: {
      const stylesMap = fromClass(delta[1]);
      for (let i = 0;i < nodes.length;i++) {
        const node = nodes[i];
        const classMap = fromClass(node.className);

        for (let j in stylesMap) if (stylesMap.hasOwnProperty(j)) {
          classMap[j] = true;
        }

        node.className = toClass(classMap);
      }

      break
    }

    case rmClassesType: {
      const classMap = fromClass(delta[1]);
      for (let i = 0;i < nodes.length;i++) {
        let node = nodes[i];
        const nodeClassMap = fromClass(node.className);

        for (let j in classMap) if (classMap.hasOwnProperty(j)) {
          delete nodeClassMap[j];
        }

        node.className = toClass(nodeClassMap);
      }

      break
    }

    case htmlType:
      return forEach(nodes, (node) => {
        node.innerHTML = '';
        return transfer(child => node.appendChild(child), getFragment(node, delta[1]))
      })

    case replaceType:
      return forEach(nodes, (node) => {
        const parent = node.parentNode;
        if (!parent) {
          return Promise.resolve()
        }

        const ref = node.nextSibling;
        parent.removeChild(node);

        return transfer(child => parent.insertBefore(child, ref), getFragment(parent, delta[1]))
      })

    case appendType:
      return forEach(nodes, (node) => {
        return transfer(child => node.appendChild(child), getFragment(node, delta[1]))
      })

    case prependType:
      return forEach(nodes, (node) => {
        const firstChild = node.firstChild;
        return transfer(child => node.insertBefore(child, firstChild), getFragment(node, delta[1]))
      })

    case insertAfterType:
      return forEach(nodes, (node) => {
        if (!node.parentNode) {
          return Promise.resolve()
        }

        return transfer(child => node.parentNode.insertBefore(child, node.nextSibling), getFragment(node.parentNode, delta[1]))
      })

    case insertBeforeType:
      return forEach(nodes, (node) => {
        if (!node.parentNode) {
          return Promise.resolve()
        }

        return transfer(child => node.parentNode.insertBefore(child, node), getFragment(node.parentNode, delta[1]))
      })

  }

  return Promise.resolve()
}

var mutex = window.witMutex = window.witMutex || (() => {
  let queued = [];
  let locked = true;

  const unlockQueue = () => {
    if (queued.length > 0) {
      queued.shift()();
    } else {
      locked = false;
    }
  };

  const mutex = {
    lock() {
      return new Promise(resolve => {
        let unlock = () => {
          unlock = () => {};
          unlockQueue();
        };

        if (!locked) {
          locked = true;
          resolve(() => unlock());
          return
        }

        queued.push(() => resolve(() => unlock()));
      })
    },
  };

  let cb = () => {
    if (window.removeEventListener) {
      window.removeEventListener('load', cb, false);
      document.removeEventListener('DOMContentLoaded', cb, false);
    } else if (window.detachEvent) {
      window.detachEvent('onload', cb);
    }

    cb = () => {};
    unlockQueue();
  };

  if (document.readyState == 'complete' || document.readyState == 'interactive') {
    unlockQueue();
  } else if (window.addEventListener) {
    window.addEventListener('load', cb, false);
    document.addEventListener('DOMContentLoaded', cb, false);
  } else if (window.attachEvent) {
    window.attachEvent('onload', cb);
  }

  return mutex
})();

var index = (delta, ...args) => {
  let rootNode, nodes;

  for (const arg of args) {
    if (arg instanceof Array) {
      nodes = arg;
    } else {
      rootNode = arg;
    }
  }

  rootNode = rootNode || document.documentElement;
  nodes = nodes || [rootNode];
  return mutex.lock().then(unlock => {
    return apply(delta, rootNode, nodes).then(unlock, unlock)
  })
};

export { addClasses, all, append, index as apply, clear, firstChild, html, insertAfter, insertBefore, lastChild, list, mutex, nextSibling, nil, one, parent, prepend, prevSibling, remove, replace, replaceAttr, rmAttr, rmClasses, rmStyles, root, setAttr, setStyles };
